div.md-typeset__scrollwrap {
    text-align: center !important;
}

body {
    counter-reset: h2;
}

h2 {
    counter-reset: h3;
}

h3 {
    counter-reset: h4;
}

h4 {
    counter-reset: h5;
}

h5 {
    counter-reset: h6;
}

article h2:before {
    content: counter(h2, decimal) ". ";
    counter-increment: h2;
}

article h3:before {
    content: counter(h2, decimal) "." counter(h3, decimal) ". ";
    counter-increment: h3;
}

article h4:before {
    content: counter(h2, decimal) "." counter(h3, decimal) "." counter(h4, decimal) ". ";
    counter-increment: h4;
}

article h5:before {
    content: counter(h2, decimal) "." counter(h3, decimal) "." counter(h4, decimal) "." counter(h5, decimal) ". ";
    counter-increment: h5;
}

article h6:before {
    content: counter(h2, decimal) "." counter(h3, decimal) "." counter(h4, decimal) "." counter(h5, decimal) "." counter(h6, decimal) ". ";
    counter-increment: h6;
}

h2.nocount:before, h3.nocount:before, h4.nocount:before, h5.nocount:before, h6.nocount:before {
    content: "";
    counter-increment: none;
}

nav.md-nav--secondary ul.md-nav__list {
    counter-reset: item
}

nav.md-nav--secondary ul.md-nav__list li a:before {
    content: counters(item, ".") " ";
    counter-increment: item;
    margin-right: .5em;
    align-self: flex-start;
}

nav.md-nav--secondary ul.md-nav__list li a {
    justify-content: flex-start;
}

.md-typeset ul li ul {
    list-style-type: circle;
}
